import threading
import logging
import select
import socket
import json
import hmac
import binascii
import os
from common.metaclasses import ServerVerifier
from common.descriptors import Port, Address
from common.utils import get_message, send_message
from common.decorators import login_required
from common.variables import MAX_CONNECTIONS, DESTINATION, SENDER, PRESENCE, TIME, USER, ACTION, ACCOUNT_NAME, \
    RESPONSE_200, RESPONSE_400, ERROR, MESSAGE, MESSAGE_TEXT, EXIT, GET_CONTACTS, RESPONSE_202, LIST_INFO, \
    REMOVE_CONTACT, USERS_REQUEST, PUBLIC_KEY_REQUEST, RESPONSE_511, DATA, RESPONSE, PUBLIC_KEY, RESPONSE_205, \
    ADD_CONTACT

# Инициализация логгера сервера
LOGGER = logging.getLogger('server')


class Server(threading.Thread):
    """
    Основной класс сервера. Принимает содинения, словари - пакеты
    от клиентов, обрабатывает поступающие сообщения.
    Работает в качестве отдельного потока.
    """
    # Дескриптор порта и ip адреса
    port = Port()
    address = Address()

    def __init__(self, listen_address, listen_port, database):
        # Параментры подключения
        self.listen_address = listen_address
        self.listen_port = listen_port
        self.database = database
        # Сокет, через который будет осуществляться работа
        self.main_socket = None
        # Устанавливаем список клиентов и очередь сообщений
        self.clients = []
        # Сокеты
        self.listen_sockets = None
        self.error_sockets = None
        # Флаг продолжения работы
        self.running = True
        # Словарь имен пользователей и соответствующие им сокеты
        self.names = dict()
        # Конструктор предка
        super().__init__()

    def run(self):
        """Метод основной цикл потока"""

        self.start_socket()

        # Основной цикл программы сервера
        while self.running:
            try:
                client, client_address = self.main_socket.accept()
            except OSError:
                pass
            else:
                LOGGER.info(f'Клиент {client_address} подключился.')
                client.settimeout(5)
                self.clients.append(client)

            recv_lst = []
            send_lst = []
            err_lst = []

            # Проверка на наличие ждущих клиентов
            try:
                if self.clients:
                    recv_lst, self.listen_sockets, self.error_sockets = select.select(
                        self.clients, self.clients, [], 0)
            except OSError as e:
                LOGGER.error(f'Ошибка сокетов: {e}')

            # Принимаем сообщения и если они есть, то заполняем словарь, если
            # ошибка - отключаем клиента
            if recv_lst:
                for client_with_msg in recv_lst:
                    try:
                        self.process_client_message(get_message(client_with_msg), client_with_msg)
                    except (OSError, json.JSONDecodeError, TypeError) as e:
                        LOGGER.debug(f'Получение данных из исключения клиента.', exc_info=e)
                        self.clients.remove(client_with_msg)

    def start_socket(self):
        """Метод инициализатор сокета."""
        LOGGER.info(
            f'Сервер запущен. Порт для подключений: {self.listen_port}.'
            f'Адрес с котрого принимаются сообщения: {self.listen_address}. '
        )
        # Готовим сокет
        transport = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        transport.bind((self.listen_address, self.listen_port))
        transport.settimeout(0.5)
        # Слушаем порт
        self.main_socket = transport
        self.main_socket.listen(MAX_CONNECTIONS)

    def remove_client(self, client):
        """
        Функция обработчик клиента с которым прервана связь.
        Ищет клиента и удаляет его из списков и базы:
        :param client:
        :return:
        """
        LOGGER.info(f'Клиент {client.getpeername()} отключился от сервера.')
        for name in self.names:
            if self.names[name] == client:
                self.database.user_logout(name)
                del self.names[name]
                break
        self.clients.remove(client)
        client.close()

    def process_message(self, message):
        """
        Функция адресной отправки сообщения определённому клиенту. Принимает словарь сообщение,
        список зарегистрированых пользователей и слушающие сокеты. Ничего не возвращает.
        """
        if message[DESTINATION] in self.names and self.names[message[DESTINATION]] in self.listen_sockets:
            try:
                send_message(self.names[message[DESTINATION]], message)
                LOGGER.info(
                    f'Отправлено сообщение пользователю {message[DESTINATION]} '
                    f'от пользователя {message[SENDER]}.')
            except OSError:
                self.remove_client(self.names[message[DESTINATION]])

        elif message[DESTINATION] in self.names and self.names[message[DESTINATION]] not in self.listen_sockets:
            LOGGER.error(
                f'Связь с пользователем {message[DESTINATION]} потеряна. '
                f'Соединение закрыто, отправка сообщения невозможна.')
            self.remove_client(self.names[message[DESTINATION]])
        else:
            LOGGER.error(
                f'Пользователь {message[DESTINATION]} не зарегистрирован на сервере, '
                f'отправка сообщения невозможна.')

    @login_required
    def process_client_message(self, message, client):
        """ Обработчик сообщений от клиентов """
        LOGGER.debug(f'Разбор сообщения от клиента : {message}')
        # Если это сообщение о присутствии, принимаем и отвечаем
        if ACTION in message and message[ACTION] == PRESENCE and TIME in message and USER in message:
            # Если сообщение о присутствии то вызываем функцию авторизации.
            self.autorized_user(message, client)

        # Если это сообщение, то добавляем его в очередь
        elif ACTION in message and message[ACTION] == MESSAGE and TIME in message and DESTINATION in message \
                and SENDER in message and MESSAGE_TEXT in message and self.names[message[SENDER]] == client:
            if message[DESTINATION] in self.names:
                self.database.process_message(message[SENDER], message[DESTINATION])
                self.process_message(message)
                try:
                    send_message(client, RESPONSE_200)
                except OSError:
                    self.remove_client(client)
            else:
                response = RESPONSE_400
                response[ERROR] = 'Пользователь не зарегистрирован.'
                try:
                    send_message(client, response)
                except OSError:
                    pass
            return

        # Если клиент выходит
        elif ACTION in message and message[ACTION] == EXIT and ACCOUNT_NAME in message and \
                self.names[message[ACCOUNT_NAME]] == client:
            self.remove_client(client)
            LOGGER.info(f'Клиент {message[ACCOUNT_NAME]} коректно вышел')

        # Если запрос контактов
        elif ACTION in message and message[ACTION] == GET_CONTACTS and USER in message and \
                self.names[message[USER]] == client:
            response = RESPONSE_202
            response[LIST_INFO] = self.database.get_contacts(message[USER])
            try:
                send_message(client, response)
            except OSError:
                self.remove_client(client)

        # Если запрос на добавление контакта
        elif ACTION in message and message[ACTION] == ADD_CONTACT and ACCOUNT_NAME in message and USER in message \
                and self.names[message[USER]] == client:
            self.database.add_contact(message[USER], message[ACCOUNT_NAME])
            try:
                send_message(client, RESPONSE_200)
            except OSError:
                self.remove_client(client)

        # Если запрос на удаление контакта
        elif ACTION in message and message[ACTION] == REMOVE_CONTACT and ACCOUNT_NAME in message and USER in message \
                and self.names[message[USER]] == client:
            self.database.remove_contact(message[USER], message[ACCOUNT_NAME])
            try:
                send_message(client, RESPONSE_200)
            except OSError:
                self.remove_client(client)

        # Если запрос известных пользователей
        elif ACTION in message and message[ACTION] == USERS_REQUEST and ACCOUNT_NAME in message \
                and self.names[message[ACCOUNT_NAME]] == client:
            response = RESPONSE_202
            response[LIST_INFO] = [user[0] for user in self.database.users_list()]
            try:
                send_message(client, response)
            except OSError:
                self.remove_client(client)

        # Если запрос публичного ключа пользователя
        elif ACTION in message and message[ACTION] == PUBLIC_KEY_REQUEST and ACCOUNT_NAME in message:
            response = RESPONSE_511
            response[DATA] = self.database.get_public_key(message[ACCOUNT_NAME])
            if response[DATA]:
                try:
                    send_message(client, response)
                except OSError:
                    self.remove_client(client)
            else:
                response = RESPONSE_400
                response[ERROR] = 'Публичный ключ не найден для данного пользователя.'
                try:
                    send_message(client, response)
                except OSError:
                    self.remove_client(client)

        # Иначе отдаём Bad request
        else:
            response = RESPONSE_400
            response[ERROR] = 'Запрос некорректен.'
            try:
                send_message(client, response)
            except OSError:
                self.remove_client(client)

    def autorized_user(self, message, client):
        """
        Функция реализующая авторизацию пользователей.
        :param message:
        :param client:
        :return:
        """

        # Если имя пользователя уже занято то возвращаем 400
        LOGGER.debug(f'Авторизация для пользователя: {message[USER]}')

        if message[USER][ACCOUNT_NAME] not in self.names.keys():
            response = RESPONSE_400
            response[ERROR] = 'Имя пользователя уже занято.'
            try:
                LOGGER.debug(f'Имя занято {response}')
                send_message(client, response)
            except OSError:
                LOGGER.debug(f'Ошибка системы')
                pass
            self.clients.remove(client)
            client.close()

        # Проверяем что пользователь зарегистрирован на сервере.
        elif not self.database.check_user(message[USER][ACCOUNT_NAME]):
            response = RESPONSE_400
            response[ERROR] = 'Пользователь не зарегистрирован.'
            try:
                LOGGER.debug(f'Неизвестное имя пользователя {response}')
                send_message(client, response)
            except OSError:
                pass
            self.clients.remove(client)
            client.close()
        else:
            LOGGER.debug('Имя пользователя корректно. Проверяем пароль.')
            # Словарь - заготовка
            message_auth = RESPONSE_511
            # Набор байтов в hex представлении
            random_str = binascii.hexlify(os.urandom(64))
            # В словарь байты нельзя, декодируем (json.dumps -> TypeError)
            message_auth[DATA] = random_str.decode('ascii')
            # Создаём хэш пароля и связки с рандомной строкой, сохраняем
            # серверную версию ключа
            hash = hmac.new(self.database.get_hash(message[USER][ACCOUNT_NAME]), random_str)
            digest = hash.digest()
            LOGGER.debug(f'Auth message = {message_auth}')
            try:
                # Обмен с клиентом
                send_message(client, message_auth)
                ans = get_message(client)
            except OSError as err:
                LOGGER.debug('Ошибка авторизации:', exc_info=err)
                client.close()
                return
            client_digest = binascii.a2b_base64(ans[DATA])
            # Если ответ клиента корректный, то сохраняем его в список
            # пользователей.
            if RESPONSE in ans and ans[RESPONSE] == 511 and hmac.compare_digest(digest, client_digest):
                self.names[message[USER][ACCOUNT_NAME]] = client
                client_ip, client_port = client.getpeername()
                try:
                    send_message(client, RESPONSE_200)
                except OSError:
                    self.remove_client(message[USER][ACCOUNT_NAME])
                # добавляем пользователя в список активных и если у него изменился открытый ключ
                # сохраняем новый
                self.database.user_login(
                    message[USER][ACCOUNT_NAME],
                    client_ip,
                    client_port,
                    message[USER][PUBLIC_KEY])
            else:
                response = RESPONSE_400
                response[ERROR] = 'Неверный пароль.'
                try:
                    send_message(client, response)
                except OSError:
                    pass
                self.clients.remove(client)
                client.close()

    def service_update_lists(self):
        """
        Функция реализующая отправку сервисного сообщения "205" клиентам.
        :return:
        """
        for client in self.names:
            try:
                send_message(self.names[client], RESPONSE_205)
            except OSError:
                self.remove_client(self.names[client])



# logging - стандартный модуль для организации логирования
import os
import sys
import logging
import logging.config
import logging.handlers
from common.variables import LOGGING_LEVEL

sys.path.append('../')

# Создаем объект форматирования:
SERVER_FORMATTER = logging.Formatter('%(asctime)s %(levelname)s %(module)s %(message)s')

# Подготовка имени файла для логирования
PATH = os.path.dirname(os.path.abspath(__file__))
#PATH = os.getcwd()
PATH = os.path.join(PATH, 'server.log')

# Создаем файловый обработчик логирования (можно задать кодировку):
SERVER_FILE_LOG = logging.handlers.TimedRotatingFileHandler(PATH, encoding='utf-8', interval=1, when='D')
SERVER_FILE_LOG.setFormatter(SERVER_FORMATTER)
steam = logging.StreamHandler(sys.stderr)
steam.setFormatter(SERVER_FORMATTER)
steam.setLevel(logging.INFO)

# Создаем объект-логгер
LOGGER = logging.getLogger('server')

# Добавляем в логгер новый обработчик событий и устанавливаем уровень логирования
LOGGER.addHandler(SERVER_FILE_LOG)
LOGGER.setLevel(LOGGING_LEVEL)
LOGGER.addHandler(steam)

# отладка
if __name__ == '__main__':
    LOGGER.critical('Критическая ошибка')
    LOGGER.error('Ошибка')
    LOGGER.debug('Отладочная информация')
    LOGGER.info('Информационное сообщение')



"""Декораторы"""

import sys
import socket
import logging
import log.server_log_config
import log.client_log_config


if sys.argv[0].find('client') == -1:
    LOGGER = logging.getLogger('server')
else:
    LOGGER = logging.getLogger('client')


def log(func):
    """Декоратор, выполняющий логирование вызовов функций."""
    def log_save(*args, **kwargs):
        res = func(*args, **kwargs)
        LOGGER.debug(f'Функция {func.__name__} была вызвана из модуля {func.__module__}'
                     f' с параметрами {args}, {kwargs}.')
        return res
    return log_save


def login_required(func):
    """
    Декоратор, проверяющий, что клиент авторизован на сервере.
    Проверяет, что передаваемый объект сокета находится в
    списке авторизованных клиентов.
    За исключением передачи словаря-запроса
    на авторизацию. Если клиент не авторизован,
    генерирует исключение TypeError
    """
    def check(*args, **kwargs):
        # проверяем, что первый аргумент - экземпляр Server
        from server.core import Server
        from common.variables import ACTION, PRESENCE
        if isinstance(args[0], Server):
            found = False
            for arg in args:
                if isinstance(arg, socket.socket):
                    # Проверяем, что данный сокет есть в списке names класса
                    for client in args[0].names:
                        if args[0].names[client] == arg:
                            found = True
            # Проверяем, что передаваемые аргументы не presence
            # сообщение. Если presense, то разрешаем
            for arg in args:
                if isinstance(arg, dict):
                    if ACTION in arg and arg[ACTION] == PRESENCE:
                        found = True
            # Если не авторизован и не сообщение начала авторизации, то
            # вызываем исключение.
            if not found:
                raise TypeError
        return func(*args, **kwargs)
    return check



"""Unit-тесты утилит"""

import os
import sys
import unittest
import json

sys.path.append(os.path.join(os.getcwd(), '..'))
from common.variables import RESPONSE, ERROR, USER, ACCOUNT_NAME, TIME, ACTION, PRESENCE, ENCODING
from common.utils import get_message, send_message


class TestSocket:
    '''
    Тестовый класс тестирования отправки и получения сообщения
    '''

    def __init__(self, test_dict):
        self.test_dict = test_dict
        self.encoded_message = None
        self.recieved_message = None

    def send(self, message_send):
        json_test_message = json.dumps(self.test_dict)
        self.encoded_message = json_test_message.encode(ENCODING)
        self.recieved_message = message_send

    def recv(self, max_len):
        json_test_message = json.dumps(self.test_dict)
        return json_test_message.encode(ENCODING)


class Tests(unittest.TestCase):
    '''
    Тестовый класс, выполняющий тестирование
    '''
    test_dict_send = {
        ACTION: PRESENCE,
        TIME: 123456.123456,
        USER: {
            ACCOUNT_NAME: 'test_user'
        }
    }
    test_dict_recv_ok = {RESPONSE: 200}
    test_dict_recv_err = {RESPONSE: 400, ERROR: 'Bad Request'}

    def test_send_msg(self):
        '''
        Тест функции отправки
        :return:
        '''
        test_socket = TestSocket(self.test_dict_send)
        send_message(test_socket, self.test_dict_send)
        # Проверяем коррректность кодирования
        self.assertEqual(test_socket.encoded_message, test_socket.recieved_message)
        # Проверяем, если на выходе не словарь
        self.assertRaises(TypeError, send_message, test_socket, 1234)

    def test_get_msg(self):
        '''
        Тест функции приема сообщения
        :return:
        '''
        test_sock_ok = TestSocket(self.test_dict_recv_ok)
        test_sock_err = TestSocket(self.test_dict_recv_err)
        self.assertEqual(get_message(test_sock_ok), self.test_dict_recv_ok)
        self.assertEqual(get_message(test_sock_err), self.test_dict_recv_err)



from PyQt5.QtWidgets import QDialog, QPushButton, QTableView
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtCore import Qt


class HistoryWindow(QDialog):
    """
    Класс - окно со статистикой пользователей
    """
    def __init__(self, database):
        super().__init__()
        self.database = database
        self.initUI()

    def initUI(self):
        # Настройки окна:
        self.setWindowTitle('Статистика пользователей')
        self.setFixedSize(600, 700)
        self.setAttribute(Qt.WA_DeleteOnClose)
        # Лист с историей
        self.history_table = QTableView(self)
        self.history_table.move(10, 10)
        self.history_table.setFixedSize(580, 620)
        # Кнопка закрытия окна
        self.close_button = QPushButton('Закрыть', self)
        self.close_button.move(250, 650)
        self.close_button.clicked.connect(self.close)

        self.create_stat_model()

    def create_stat_model(self):
        """
        Функция реализующая заполнение таблицы
        с историей (статистикой) сообщений
        """
        # Список записей из базы
        history_list = self.database.message_history()
        # Объект модели данных:
        list = QStandardItemModel()
        list.setHorizontalHeaderLabels(
            ['Имя пользователя', 'Дата последнего входа', 'Отправлено сообщений', 'Принято сообщений'])
        for row in history_list:
            user, last_login, send, recieve = row
            user = QStandardItem(user)
            user.setEditable(False)
            last_login = QStandardItem(str(last_login.replace(microsecond=0)))
            last_login.setEditable(False)
            send = QStandardItem(str(send))
            send.setEditable(False)
            recieve = QStandardItem(str(recieve))
            recieve.setEditable(False)
            list.appendRow([user, last_login, send, recieve])
        self.history_table.setModel(list)
        self.history_table.resizeColumnsToContents()
        self.history_table.resizeRowsToContents()



"""Unit-тесты сервера"""

import os
import sys
import unittest

sys.path.append(os.path.join(os.getcwd(), '..'))

from common.variables import RESPONSE, ERROR, USER, ACCOUNT_NAME, TIME, ACTION, PRESENCE, RESPONSE_DEFAULT_IP_ADDRESS
from server import process_client_message


class TestServer(unittest.TestCase):
    '''
    Класс с тестами
    '''
    err_dict = {
        RESPONSE_DEFAULT_IP_ADDRESS: 400,
        ERROR: 'Bad Request'
    }
    ok_dict = {RESPONSE: 200}

    def test_ok_check(self):
        """Проверка запроса на корректность"""
        self.assertEqual(process_client_message(
            {ACTION: PRESENCE, TIME: 1.1, USER: {ACCOUNT_NAME: 'Guest'}}), self.ok_dict)

    def test_no_action(self):
        """Ошибка если нет действия"""
        self.assertEqual(process_client_message({TIME: '1.1', USER: {ACCOUNT_NAME: 'Guest'}}), self.err_dict)

    def test_no_user(self):
        """Тест если нет пользователя"""
        self.assertEqual(process_client_message({ACTION: PRESENCE, TIME: '1.1'}), self.err_dict)

    def test_wrong_action(self):
        """Ошибка если неизвестное действие"""
        self.assertEqual(process_client_message(
            {ACTION: 'Wrong', TIME: '1.1', USER: {ACCOUNT_NAME: 'Guest'}}), self.err_dict)

    def test_unknown_user(self):
        """Проверяем если пользователь не Guest"""
        self.assertEqual(process_client_message(
            {ACTION: PRESENCE, TIME: 1.1, USER: {ACCOUNT_NAME: 'Unknown_User'}}), self.err_dict)

    def test_no_time(self):
        """Ошибка, если  запрос не содержит штампа времени"""
        self.assertEqual(process_client_message({ACTION: PRESENCE, USER: {ACCOUNT_NAME: 'Guest'}}), self.err_dict)



import logging
import ipaddress
import sys

if sys.argv[0].find('client') == -1:
    LOGGER = logging.getLogger('server')
else:
    LOGGER = logging.getLogger('client')


class Port:
    """
    Класс - дескриптор для номера порта.
    Позволяет использовать только порты с 1023 по 65536.
    При попытке установить неподходящий номер порта генерирует исключение.
    """
    def __set__(self, instance, value):
        if value < 1024 or value > 65535:
            LOGGER.critical(
                f'Запуск сервера с параметром порта {value}.'
                f'В качестве порта может быть указано только число в диапазоне от 1024 до 65535.')
            sys.exit(1)
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name


class Address:
    """
    Класс - дескриптор для ip-адреса.
    Проверяет адрес на корректность.
    При попытке ввода невалидного ip-адреса генерирует исключение.
    """
    def __set__(self, instance, value):
        if value:
            try:
                ipv4 = ipaddress.ip_address(value)
            except ValueError as e:
                LOGGER.critical(f'Некорректно указан ip адрес {value}.')
                sys.exit(1)
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name



import sys
from PyQt5.QtWidgets import QApplication, QMainWindow, QAction, qApp,  QLabel, QTableView
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtCore import QTimer
from server.stat_window import HistoryWindow
from server.config_window import ConfigWindow
from server.add_user import RegisterUser
from server.remove_user import DeleteUser



class MainWindow(QMainWindow):
    """ Класс - основное окно сервера. """
    def __init__(self, database, server, config):
        super().__init__()
        # База данных сервера
        self.database = database

        self.server_thread = server
        self.config = config

        # кнопка выход
        self.exitAction = QAction('Выход', self)
        self.exitAction.setShortcut('Ctrl+X')
        self.exitAction.triggered.connect(qApp.quit)
        # кнопка настройки сервера
        self.config_button = QAction('Настройки сервера', self)
        # кнопка обновления списка клиентов
        self.refresh_button = QAction('Обновление списка', self)
        # кнопка истории сообщений
        self.show_history_button = QAction('История пользователей', self)

        # Кнопка регистрации пользователя
        self.register_button = QAction('Регистрация пользователя', self)

        # Кнопка удаления пользователя
        self.remove_button = QAction('Удаление пользователя', self)

        # Статусбар
        self.statusBar()
        self.statusBar().showMessage('Сервер запущен')
        # Тулбар
        self.toolbar = self.addToolBar('MainBar')
        self.toolbar.addAction(self.refresh_button)
        self.toolbar.addAction(self.show_history_button)
        self.toolbar.addAction(self.register_button)
        self.toolbar.addAction(self.remove_button)
        self.toolbar.addAction(self.config_button)
        self.toolbar.addAction(self.exitAction)

        # настройка вида основного окна
        self.setFixedSize(800, 600)
        self.setWindowTitle('Message server')

        self.label = QLabel('Список подключённых клиентов:', self)
        self.label.setFixedSize(240, 15)
        self.label.move(10, 25)
        # Окно со списком подключённых клиентов.
        self.active_clients_table = QTableView(self)
        self.active_clients_table.move(10, 45)
        self.active_clients_table.setFixedSize(780, 400)

        # Таймер, обновляющий список клиентов 1 раз в секунду
        self.timer = QTimer()
        self.timer.timeout.connect(self.create_gui_model)
        self.timer.start(1000)

        # Связываем кнопки с процедурами
        self.refresh_button.triggered.connect(self.create_gui_model)
        self.show_history_button.triggered.connect(self.show_statistics)
        self.config_button.triggered.connect(self.server_config)
        self.register_button.triggered.connect(self.reg_user)
        self.remove_button.triggered.connect(self.rem_user)

        # Отображаем окно.
        self.show()

    def create_gui_model(self):
        """Функция заполняющая таблицу активных пользователей."""
        users_list_active = self.database.active_users_list()
        list = QStandardItemModel()
        list.setHorizontalHeaderLabels(['Имя пользователя', 'IP адрес', 'Порт', 'время подключения'])
        for row in users_list_active:
            user, ip, port, time = row
            user = QStandardItem(user)
            user.setEditable(False)
            ip = QStandardItem(ip)
            ip.setEditable(False)
            port = QStandardItem(str(port))
            port.setEditable(False)
            time = QStandardItem(str(time.replace(microsecond=0)))
            time.setEditable(False)
            list.appendRow([user, ip, port, time])
        self.active_clients_table.setModel(list)
        self.active_clients_table.resizeColumnsToContents()
        self.active_clients_table.resizeRowsToContents()

    def show_statistics(self):
        """Метод создающий окно со статистикой клиентов."""
        global history_window
        history_window = HistoryWindow(self.database)
        history_window.show()

    def server_config(self):
        """Метод создающий окно с настройками сервера."""
        global config_window
        # Создаём окно и заносим в него текущие параметры
        config_window = ConfigWindow(self.config)

    def reg_user(self):
        """Метод создающий окно регистрации пользователя"""
        global reg_window
        reg_window = RegisterUser(self.database, self.server_thread)
        reg_window.show()

    def rem_user(self):
        """"Метод создающий окно удаления пользователя."""
        global rem_window
        rem_window = DeleteUser(self.database, self.server_thread)
        rem_window.show()



import datetime

from sqlalchemy import create_engine, Table, Column, Integer, String, MetaData, ForeignKey, DateTime, Text
from sqlalchemy.orm import mapper, sessionmaker
from sqlalchemy.sql import default_comparator


class ServerDatabase:
    '''
    Класс - оболочка для работы с базой данных сервера.
    Использует SQLite базу данных, реализован с помощью
    SQLAlchemy ORM и используется классический подход.
    '''

    class AllUsers:
        """Класс - отображение таблицы всех пользователей"""

        def __init__(self, username, passwd_hash):
            self.name = username
            self.last_login = datetime.datetime.now()
            self.passwd_hash = passwd_hash
            self.public_key = None
            self.id = None

    class ActiveUsers:
        """Класс - отображение таблицы активных пользователей"""

        def __init__(self, user_id, ip_address, port, login_time):
            self.user = user_id
            self.ip_address = ip_address
            self.port = port
            self.login_time = login_time
            self.id = None

    class LoginHistory:
        """Класс - отображение таблицы истории логирования"""

        def __init__(self, name, date, ip, port):
            self.id = None
            self.name = name
            self.date_time = date
            self.ip = ip
            self.port = port

    class UserContactList:
        """Класс - отображение таблицы контактов пользователей"""

        def __init__(self, user, contact):
            self.id = None
            self.user = user
            self.contact = contact

    class UsersHistory:
        """Класс отображение таблицы истории действий"""

        def __init__(self, user):
            self.id = None
            self.user = user
            self.send = 0
            self.accepted = 0

    def __init__(self, path):
        # Создаём движок базы данных
        print(path)
        self.database_engine = create_engine(
            f'sqlite:///{path}',
            echo=False,
            pool_recycle=7200,
            connect_args={
                'check_same_thread': False})
        self.metadata = MetaData()
        # Создаём таблицу пользователей
        users_table = Table('Users', self.metadata,
                            Column('id', Integer, primary_key=True),
                            Column('name', String, unique=True),
                            Column('last_login', DateTime),
                            Column('passwd_hash', String),
                            Column('public_key', Text)
                            )
        # Создаём таблицу активных пользователей
        active_users_table = Table(
            'Active_users', self.metadata, Column(
                'id', Integer, primary_key=True), Column(
                'user', ForeignKey('Users.id'), unique=True), Column(
                'ip_address', String), Column(
                    'port', Integer), Column(
                        'login_time', DateTime))
        # Создаём таблицу истории входов
        user_login_history = Table('Login_history', self.metadata,
                                   Column('id', Integer, primary_key=True),
                                   Column('name', ForeignKey('Users.id')),
                                   Column('date_time', DateTime),
                                   Column('ip', String),
                                   Column('port', String)
                                   )
        # Создаём таблицу контактов пользователей
        contacts = Table('Contacts', self.metadata,
                         Column('id', Integer, primary_key=True),
                         Column('user', ForeignKey('Users.id')),
                         Column('contact', ForeignKey('Users.id'))
                         )

        # Создаём таблицу истории пользователей
        users_history_table = Table('History', self.metadata,
                                    Column('id', Integer, primary_key=True),
                                    Column('user', ForeignKey('Users.id')),
                                    Column('send', Integer),
                                    Column('accepted', Integer)
                                    )

        # Создание таблиц
        self.metadata.create_all(self.database_engine)
        # Создание отображений
        mapper(self.AllUsers, users_table)
        mapper(self.ActiveUsers, active_users_table)
        mapper(self.LoginHistory, user_login_history)
        mapper(self.UserContactList, contacts)
        mapper(self.UsersHistory, users_history_table)
        # Создание сессии
        Session = sessionmaker(bind=self.database_engine)
        self.session = Session()

        self.session.query(self.ActiveUsers).delete()
        self.session.commit()

    # При входе пользователя проверяем если пользователь есть, записываем в базу факт входа,
    # если нет, то создаем нового пользователя в БД и фиксируем факт входа
    def user_login(self, username, ip_address, port, key):
        """
        Метод выполняющийся при входе пользователя, записывает в базу факт входа
        Обновляет открытый ключ пользователя при его изменении.
        """
        is_user_exist = self.session.query(
            self.AllUsers).filter_by(
            name=username)
        # Если имя пользователя уже присутствует в таблице, обновляем время последнего входа
        # и проверяем корректность ключа
        if is_user_exist.count():
            user = is_user_exist.first()
            user.last_login = datetime.datetime.now()
            if user.public_key != key:
                user.public_key = key
        else:
            raise ValueError('Пользователь не зарегистрирован')

        # Создаем запись в таблицу активных пользователей о факте входа
        new_active_user = self.ActiveUsers(
            user.id, ip_address, port, datetime.datetime.now())
        self.session.add(new_active_user)

        history = self.LoginHistory(
            user.id, datetime.datetime.now(), ip_address, port)
        self.session.add(history)

        self.session.commit()

    # Если пользователь вышел, удаляем его из таблицы активных пользователей
    def user_logout(self, username):
        """Метод фиксирующий отключения пользователя."""
        user = self.session.query(
            self.AllUsers).filter_by(
            name=username).first()

        self.session.query(self.ActiveUsers).filter_by(user=user.id).delete()
        self.session.commit()

    # Функция регистрации пользователя в базе
    def add_user(self, name, passwd_hash):
        """
        Метод регистрации пользователя.
        Принимает имя и хэш пароля, создаёт запись в таблице статистики.
        """
        user_row = self.AllUsers(name, passwd_hash)
        self.session.add(user_row)
        self.session.commit()
        history_row = self.UsersHistory(user_row.id)
        self.session.add(history_row)
        self.session.commit()

    # Функция удаления пользователя из базы
    def remove_user(self, name):
        """Метод удаляющий пользователя из базы."""
        user = self.session.query(self.AllUsers).filter_by(name=name).first()
        self.session.query(self.ActiveUsers).filter_by(user=user.id).delete()
        self.session.query(self.LoginHistory).filter_by(name=user.id).delete()
        self.session.query(
            self.UserContactList).filter_by(
            user=user.id).delete()
        self.session.query(
            self.UserContactList).filter_by(
            contact=user.id).delete()
        self.session.query(self.UsersHistory).filter_by(user=user.id).delete()
        self.session.query(self.AllUsers).filter_by(name=name).delete()
        self.session.commit()

    # Функция получения хэша пароля
    def get_hash(self, name):
        """Метод получения хэша пароля пользователя."""
        user = self.session.query(self.AllUsers).filter_by(name=name).first()
        return user.passwd_hash

    # Функция получения публичного ключа пользователя
    def get_public_key(self, name):
        """Метод получения публичного ключа пользователя."""
        user = self.session.query(self.AllUsers).filter_by(name=name).first()
        return user.public_key

    # Функция проверки существования пользователя
    def check_user(self, name):
        """Метод проверяющий существование пользователя."""
        if self.session.query(self.AllUsers).filter_by(name=name).count():
            return True
        else:
            return False

    # Функция фиксирует передачу сообщения и делает соответствующие отметки в
    # БД
    def process_message(self, sender, recipient):
        """Метод записывающий в таблицу статистики факт передачи сообщения."""
        # Получение ID отправителя и получателя
        sender = self.session.query(
            self.AllUsers).filter_by(
            name=sender).first().id
        recipient = self.session.query(
            self.AllUsers).filter_by(
            name=recipient).first().id
        # Запрашиваем строки из истории и увеличиваем счётчики
        sender_row = self.session.query(
            self.UsersHistory).filter_by(
            user=sender).first()
        sender_row.send += 1
        recipient_row = self.session.query(
            self.UsersHistory).filter_by(
            user=recipient).first()
        recipient_row.accepted += 1

        self.session.commit()

    # Функция добавления контактов пользователя
    def add_contact(self, user, contact):
        """Метод добавления контакта для пользователя."""
        user = self.session.query(self.AllUsers).filter_by(name=user).first()
        contact = self.session.query(
            self.AllUsers).filter_by(
            name=contact).first()

        if not contact or self.session.query(
                self.UserContactList).filter_by(
                user=user.id,
                contact=contact.id).count():
            return

        contact_list = self.UserContactList(user.id, contact.id)
        self.session.add(contact_list)
        self.session.commit()

    # Функция удаления контактов пользователя
    def remove_contact(self, user, contact):
        """Метод удаления контакта пользователя."""
        user = self.session.query(self.AllUsers).filter_by(name=user).first()
        contact = self.session.query(
            self.AllUsers).filter_by(
            name=contact).first()
        # Проверяем что контакт может существовать
        if not contact:
            return
        # Удаляем
        self.session.query(self.UserContactList).filter(
            self.UserContactList.user == user.id,
            self.UserContactList.contact == contact.id).delete()
        self.session.commit()

    # Функция возвращает список известных пользователей со временем последнего
    # входа.
    def users_list(self):
        """Метод возвращающий список известных пользователей со временем последнего входа."""
        query = self.session.query(
            self.AllUsers.name,
            self.AllUsers.last_login)
        return query.all()

    # Функция возвращает список активных пользователей
    def active_users_list(self):
        """Метод возвращающий список активных пользователей."""
        query = self.session.query(self.AllUsers.name,
                                   self.ActiveUsers.ip_address,
                                   self.ActiveUsers.port,
                                   self.ActiveUsers.login_time
                                   ).join(self.AllUsers)
        return query.all()

    # Функция возвращающаяя историю входов по пользователю или всем
    # пользователям
    def login_history(self, username=None):
        """Метод возвращающий историю входов"""
        query = self.session.query(self.AllUsers.name,
                                   self.LoginHistory.date_time,
                                   self.LoginHistory.ip,
                                   self.LoginHistory.port
                                   ).join(self.AllUsers)
        if username:
            query = query.filter(self.AllUsers.name == username)
        return query.all()

    # Функция возвращает список контактов пользователя.
    def get_contacts(self, username):
        """Метод возвращающий список контактов пользователя."""
        # Запрашивааем указанного пользователя
        user = self.session.query(self.AllUsers).filter_by(name=username).one()
        # Запрашиваем его список контактов
        query = self.session.query(
            self.UserContactList,
            self.AllUsers.name). filter_by(
            user=user.id). join(
            self.AllUsers,
            self.UserContactList.contact == self.AllUsers.id)

        return [contact[1] for contact in query.all()]

    # Функция возвращает количество переданных и полученных сообщений
    def message_history(self):
        """Метод возвращающий статистику сообщений."""
        query = self.session.query(
            self.AllUsers.name,
            self.AllUsers.last_login,
            self.UsersHistory.send,
            self.UsersHistory.accepted).join(
            self.AllUsers)
        return query.all()


if __name__ == '__main__':
    test_db = ServerDatabase()
    test_db.user_login('test_user_1', '192.168.32.17', 7779)
    test_db.user_login('test_user_2', '192.168.38.123', 7777)

    print(test_db.users_list())
    print(test_db.active_users_list())

    test_db.user_logout('test_user_1')
    print(test_db.active_users_list())

    test_db.login_history('test_user_1')
    print(test_db.users_list())
    test_db.add_contact('test_user_2', 'test_user_1')
    test_db.add_contact('test_user_2', 'test_user_3')
    test_db.add_contact('test_user_1', 'test_user_4')
    print(test_db.remove_contact('test_user_1'))
    test_db.remove_contact('test_user_1', 'test_user_4')
    print(test_db.remove_contact('test_user_1'))
    test_db.process_message('test_user_1', 'test_user_2')
    print(test_db.message_history())
    print()
    print(test_db.remove_contact('test_user_2'))



import configparser
import os
import sys
import logging
import argparse
import log.server_log_config
from common.decorators import log
from common.variables import DEFAULT_PORT
from server.core import Server
from server.server_database import ServerDatabase
from server.main_window import MainWindow
from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import Qt

# Инициализация логгера сервера
LOGGER = logging.getLogger('server')


@log
def create_arg_parser(default_address, default_port):
    """ Парсер аргументов коммандной строки """
    LOGGER.debug(
        f'Инициализация парсера аргументов коммандной строки: {sys.argv}')
    parser = argparse.ArgumentParser()
    parser.add_argument('-a', '--addr', default=default_address, nargs='?')
    parser.add_argument(
        '-p',
        '--port',
        default=default_port,
        type=int,
        nargs='?')
    parser.add_argument('--no_gui', action='store_true')
    namespace = parser.parse_args(sys.argv[1:])
    listen_address = namespace.addr
    listen_port = namespace.port
    gui_flag = namespace.no_gui
    return listen_address, listen_port, gui_flag


@log
def config_load():
    """ Парсер конфигурационного ini файла. """
    config = configparser.ConfigParser()

    #dir_path = os.path.dirname(os.path.realpath(__file__))
    dir_path = os.getcwd()
    config.read(f"{dir_path}/{'server.ini'}")
    # Если конфиг файл загружен правильно, запускаемся, иначе конфиг по
    # умолчанию.
    if 'SETTINGS' in config:
        return config
    else:
        config.add_section('SETTINGS')
        config.set('SETTINGS', 'Default_port', str(DEFAULT_PORT))
        config.set('SETTINGS', 'Listen_Address', '')
        config.set('SETTINGS', 'Database_path', '')
        config.set('SETTINGS', 'Database_file', 'server_base.db3')
        return config


@log
def main():
    """ Основная функция сервера"""
    # Загружаем файл конфигурации
    config = config_load()

    # Проверяем параметры командной строки
    listen_address, listen_port, gui_flag = create_arg_parser(
        config['SETTINGS']['Listen_Address'], config['SETTINGS']['Default_port'])

    database = ServerDatabase(os.path.join(
        config['SETTINGS']['Database_path'],
        config['SETTINGS']['Database_file']))

    # Создание экземпляра класса - сервера и его запуск:
    server = Server(listen_address, listen_port, database)
    server.daemon = True
    server.start()

    if gui_flag:
        while True:
            command = input('Введите exit для завершения работы сервера.')
            if command == 'exit':
                # Если выход, то завршаем основной цикл сервера.
                server.running = False
                server.join()
                break
    else:
        # Графическое окружение сервера
        server_app = QApplication(sys.argv)
        server_app.setAttribute(Qt.AA_DisableWindowContextHelpButton)
        main_window = MainWindow(database, server, config)
        # Запуск графического интерфейса
        server_app.exec_()
        # По закрытию окон останавливаем обработчик сообщений
        server.running = False


if __name__ == '__main__':
    main()
